;nyquist plug-in
;version 4
;type process
;name "Accesstrum"
;action "Gathering Data..."
;info "by Robert J. Haenggi\nReleased under GPL v2.\n"
;;
;; rjh-accesstrum.ny
; Demo at https://www.dropbox.com/s/pd4altu783nvbof/accessible_spectrum.mp3?dl=0
;
;preview enabled
;
;control win-type "Window" choice "Rectangular,Hann,Hamming,Poisson,Hann-Poisson,Flat-Top,Gauss" 0
;control fft-size "Size" choice "64,128,256,512,1024,2048,4096,8192,16384,32768,65536" 4
;control contrast "Contrast" real "" 1.0 0.0 10.0
;control hold "extent   and hold Peaks" int "" 1 1 50
;control slopes "Derivatives" int "" 0 0 10
;
;(defconstant pi 3.14159265358979323846264338327)
;
(defun f-to-pitch (frequency)
  (if (< frequency (step-to-hz c0)) (format nil  "--\t--\t--")
     (let* ((pitch (hz-to-step frequency))
       (half-steps (round pitch))
       (octave (1- (/ half-steps 12)))
       (cents (round (* 100 (- pitch half-steps))))
       (pitch-name (nth (rem half-steps 12) '("C" "C\#" "D" "D\#" "E" "F" "F\#" "G" "G\#" "A" "A\#" "B"))))
      ; Note: cents means in this context percent away from Equal Temp.
      (format nil "~a\t~a\t~a~a %"
         pitch-name octave (if (plusp cents) "+" "") cents))))
; Windows
(defun poisson (tau &aux (array (make-array *size*)))
(dotimes (n *size* (snd-from-array 0 *size* array))
(setf (aref array n) (exp (- (* (abs (- n(/ (1+ *size*) 2))) 
(/ 1.0 tau)))))))
;;
(defun han (&aux (array (make-array *size*))) 
  (do* ((i 0 (1+ i)) (w 0 (/ (float i) *size*)) (sum 0)) 
     ((= i *size*) 
      (mult (/ *size* sum); (/ 2 (aref array (/ *size* 2))) 
            (snd-from-array 0 *size*  array)))
     (setf (aref array i) 
           (- 0.5 (* 0.5  (cos(* 2 pi w)))))
(setf sum (+ sum (aref array i))))) 
;;
(defun flat-top (&aux (array (make-array *size*)))
(do* ((i 0 (1+ i)) (w 0 (/ (float i) *size*)) (sum 0)) 
     ((= i *size*) 
      (mult (/ *size* sum); (/ 2 (aref array (/ *size* 2))) 
            (snd-from-array 0 *size* array)))
     (setf (aref array i) 
           (- 0.21557895
              (* 0.41663158 (cos(* 2 pi w)))
              (* -0.277263158 (cos (* 4 pi w)))
              (* 0.083578947 (cos(* 6 pi w))) 
              (* -0.006947368(cos(* 8 pi w)))))
(setf sum (+ sum (aref array i)))))
;; quadratic interpolation with sounds
(defun quadratic (s)
  (let* ((y-1 s) 
        (y0 (extract-abs (/ 1 *sound-srate*) 3600 s)) 
        (y1 (extract-abs (/ 2 *sound-srate*) 3600 s))
        (pos (clip (mult  (diff  Y1  Y-1)
          (recip (mult  2 (diff (mult  2 Y0) (sim Y1 Y-1))))) 1)))
; return estimated peak 
(s-abs (sum y0 (mult  -0.25 pos (diff  y-1 y1))))))
;;
(defun power-spectrum (frame)
  (setf (aref frame 0) (/ (aref frame 0) 2.0)) 
  (setf (aref frame (1- *size*)) (/ (aref frame (1- *size*)) 2.0))
  (let* ((snd (snd-from-array 0 2 frame))
         (zero (snd-from-array 0 2 #(0))))
   (mult 2 (snd-avg (seq (cue zero) (cue (mult (/ (expt (/ *size* 2.0) 2.0)) snd snd))) 2 2 op-average))))
;;
(defun spectral-flatness (spectrum)
  (let* (
     (n (snd-length spectrum ny:all))
     (sr (snd-srate spectrum))
     (nozero (s-max (snd-const 1e-20 0 sr (/ n sr)) spectrum))
     (geometric (exp (snd-fetch (snd-avg (s-log nozero) n n 1))))      (arithmetic (snd-fetch (snd-avg spectrum n n 1))))
   (/ geometric arithmetic)))
;;
(set-sound-srate *sound-srate*)
(setf *size* (truncate (expt 2.0 (+ fft-size 6.0))))
(setf tau (/ *size* 15.0 0.23026))
(setf *win* (case win-type
  (1 (han))
  (3 (poisson tau))
  (4 (mult (han) (poisson tau)))
  (5 (flat-top))
  (t nil)))
(let* (freq 
(half (/ *size* 2))
(no-of-frames (print (1- (/ (truncate len) half))))
(norm (/ 1.0 no-of-frames)) 
(chirp (snd-pwl 0 1.0 (list 0 0.0 half (/ *sound-srate* 2) (1+ half)))) 
(spec 
(do* ((i 0 (1+ i))
 (s-sum (cue (snd-const 0 0 1.0 (1+ half))))
(s-error (cue (snd-const 0 0 1.0 (1+ half))))
(s-total (cue (snd-const 0 0 1.0 (1+ half))))
(s-y (cue (snd-const 0 0 1.0 (1+ half)))))
 ((> i no-of-frames) (s-sqrt s-sum))
  (setf s-y  (diff (scale norm
     (cue (power-spectrum (snd-fft *track* *size* half *win*)))) s-error))
  (setf s-total  (sim s-sum s-y))
  (setf s-error (diff (diff s-total s-sum) s-y))
  (snd-length s-error ny:all)
  (setf s-sum s-total)))
(log-spec (snd-samples (linear-to-db spec) ny:all))
(smoothed (extract-abs 0 (1+ half)  (snd-avg (seq (cue (snd-const 0 0 1 (/ hold 2))) (cue spec)) hold 1 op-peak))))
;; header
(Format t "Frequency\tLevel\tName\tOctave\tOffset\n")
;; values
(do ((i 0 (1+ i))) ((= i (1+ half)))
(setf freq (* i (/ *sound-srate* *size*)))
(format t "~a\t~a\t~a~%"
 freq (aref log-spec i) (f-to-pitch freq)))
; without that, we get a crash, why that?
 (print (snd-samples smoothed 1) (make-string-output-stream))
 (setf chirp (scale-srate chirp 50))
 (setf spec (scale-srate 
(if (= contrast 0.0) (diff (snd-avg spec 64 1 op-peak) spec) 
(snd-exp (mult contrast  (snd-log smoothed)))) 
50.0)) 
(dotimes (i slopes)
(setf spec (sum spec (mult (/ -0.05 *sound-srate*) (slope spec)))))
;(setf slen (* 50 (1+ half)))
;(setf spec (sound-srate-abs 50 (sound-warp   (integrate (snd-pwl 0 50.0 (list 0 0.1   slen 1.0 (1+ slen)))) spec 50)))
(setf peak (peak spec ny:all))
(setf spec (scale (/ .99 peak) spec))
(setf output (mult spec (hzosc chirp)))
output)